@@ -76,6 +76,10 @@ def write_crc32(encoder, datum):
 def write_fixed(encoder, datum, schema, named_schemas, fname):
     """Fixed instances are encoded using the number of bytes declared in the
     schema."""
+    if len(datum) != schema["size"]:
+        raise ValueError(
+            f"data of length {len(datum)} does not match schema size: {schema}"
+        )
     encoder.write_fixed(datum)
 
 
@@ -155,15 +159,49 @@ def write_union(encoder, datum, schema, named_schemas, fname):
     else:
         pytype = type(datum)
         most_fields = -1
+
+        # All of Python's floating point values are doubles, so to
+        # avoid loss of precision, we should always prefer 'double'
+        # if we are forced to choose between float and double.
+        #
+        # If 'double' comes before 'float' in the union, then we'll immediately
+        # choose it, and don't need to worry. But if 'float' comes before
+        # 'double', we don't want to pick it.
+        #
+        # So, if we ever see 'float', we skim through the rest of the options,
+        # just to see if 'double' is a possibility, because we'd prefer it.
+        could_be_float = False
+
         for index, candidate in enumerate(schema):
+
+            if could_be_float:
+                if extract_record_type(candidate) == "double":
+                    best_match_index = index
+                    break
+                else:
+                    # Nothing except "double" is even worth considering.
+                    continue
+
             if _validate(datum, candidate, named_schemas, raise_errors=False):
-                if extract_record_type(candidate) == "record":
+                record_type = extract_record_type(candidate)
+                if record_type == "record":
+                    logical_type = extract_logical_type(candidate)
+                    if logical_type:
+                        prepare = LOGICAL_WRITERS.get(logical_type)
+                        if prepare:
+                            datum = prepare(datum, candidate)
+
                     candidate_fields = set(f["name"] for f in candidate["fields"])
                     datum_fields = set(datum)
                     fields = len(candidate_fields.intersection(datum_fields))
                     if fields > most_fields:
                         best_match_index = index
                         most_fields = fields
+                elif record_type == "float":
+                    best_match_index = index
+                    # Continue in the loop, because it's possible that there's
+                    # another candidate which has record type 'double'
+                    could_be_float = True
                 else:
                     best_match_index = index
                     break
@@ -359,7 +397,7 @@ else:
 class GenericWriter:
     def __init__(self, schema, metadata=None, validator=None):
         self._named_schemas = {}
-        self.schema = parse_schema(schema, _named_schemas=self._named_schemas)
+        self.schema = parse_schema(schema, self._named_schemas)
         self.validate_fn = _validate if validator is True else validator
         self.metadata = metadata or {}
 
@@ -636,13 +674,13 @@ def schemaless_writer(fo, schema, record):
     Example::
 
         parsed_schema = fastavro.parse_schema(schema)
-        with open('file.avro', 'rb') as fp:
+        with open('file', 'rb') as fp:
             fastavro.schemaless_writer(fp, parsed_schema, record)
 
     Note: The ``schemaless_writer`` can only write a single record.
     """
     named_schemas = {}
-    schema = parse_schema(schema, _named_schemas=named_schemas)
+    schema = parse_schema(schema, named_schemas)
 
     encoder = BinaryEncoder(fo)
     write_data(encoder, record, schema, named_schemas, "")
